# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import os
import json
import ntpath
import hashlib
import zipfile
from io import BytesIO
from typing import Callable


from pefile import MACHINE_TYPE
import jsonschema
import jsonschema.exceptions

import speakeasy
import speakeasy.winenv.arch as _arch
from speakeasy import PeFile
from speakeasy import Win32Emulator
from speakeasy import WinKernelEmulator

from speakeasy.errors import SpeakeasyError, ConfigError, NotSupportedError


class Speakeasy(object):
    """
    Wrapper class for invoking the speakeasy emulators
    """

    def check_init(func):

        """Wrapper to make sure the emulator is initialized"""

        def wrap(self, *args, **kwargs):
            if not self.emu:
                raise SpeakeasyError('Emulator not initialized')
            return func(self, *args, **kwargs)
        return wrap

    def __init__(self, config=None, logger=None, argv=[], debug=False, exit_event=None):

        self.logger = logger
        self._init_config(config)
        self.emu = None
        self.api_hooks = []
        self.code_hooks = []
        self.dyn_code_hooks = []
        self.invalid_insn_hooks = []
        self.mem_read_hooks = []
        self.argv = argv
        self.exit_event = exit_event
        self.debug = debug
        self.loaded_bins = []
        self.mem_write_hooks = []
        self.mem_invalid_hooks = []
        self.interrupt_hooks = []
        self.mem_map_hooks = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        del self

    def _init_config(self, config: dict) -> None:
        """
        Init the emulator config
        args:
            config: The configuration data to set up the emulator environment
                    If none is supplied, a default config is used.
        return:
            None
        """
        if not config:
            config_path = os.path.join(os.path.dirname(speakeasy.__file__),
                                       'configs', 'default.json')
            with open(config_path, 'r') as f:
                self.config = json.load(f)
        else:
            self.config = config

        try:
            validate_config(self.config)
        except jsonschema.exceptions.SchemaError as err:
            if self.logger:
                self.logger.exception('Invalid config schema: %s', str(err))
            raise ConfigError('Invalid config schema')
        except jsonschema.exceptions.ValidationError as err:
            if self.logger:
                self.logger.exception('Invalid config: %s', str(err))
            raise ConfigError('Invalid config')

    def _init_emulator(self, path=None, data=None, is_raw_code=False) -> None:
        """
        Based on the PE metadata, use the appropriate emulator. That is,
        a user mode emulator vs a kernel mode emulator

        """
        if not is_raw_code:
            pe = PeFile(path=path, data=data)
            # Get the machine type we only support x86/x64 atm
            mach = MACHINE_TYPE[pe.FILE_HEADER.Machine].split('_')[-1:][0].lower()
            if mach not in ('amd64', 'i386'):
                raise SpeakeasyError('Unsupported architecture: %s' % mach)

            if pe.is_dotnet():
                raise NotSupportedError('.NET assemblies are not currently supported')

            if pe.is_driver():
                self.emu = WinKernelEmulator(config=self.config, logger=self.logger,
                                             debug=self.debug, exit_event=self.exit_event)
            else:
                self.emu = Win32Emulator(config=self.config, logger=self.logger, argv=self.argv,
                                         debug=self.debug, exit_event=self.exit_event)
        else:
            self.emu = Win32Emulator(config=self.config, logger=self.logger, argv=self.argv,
                                     debug=self.debug, exit_event=self.exit_event)

    def _init_hooks(self) -> None:
        """
        Lazily add hooks if users added them early before emulator engine was instantiated
        """
        # Add any configured hooks here
        while self.api_hooks:
            h = self.api_hooks.pop(0)
            cb, mod, func, argc, cconv = h
            self.add_api_hook(cb, mod, func, argc, cconv)
        while self.code_hooks:
            h = self.code_hooks.pop(0)
            cb, begin, end, ctx = h
            self.add_code_hook(cb, begin, end, ctx)
        while self.dyn_code_hooks:
            h = self.dyn_code_hooks.pop(0)
            cb, ctx = h
            self.add_dyn_code_hook(cb, ctx)
        while self.invalid_insn_hooks:
            h = self.invalid_insn_hooks.pop(0)
            cb, ctx = h
            self.add_invalid_instruction_hook(cb, ctx)
        while self.mem_read_hooks:
            h = self.mem_read_hooks.pop(0)
            cb, begin, end = h
            self.add_mem_read_hook(cb, begin, end)
        while self.mem_write_hooks:
            h = self.mem_write_hooks.pop(0)
            cb, begin, end = h
            self.add_mem_write_hook(cb, begin, end)
        while self.mem_invalid_hooks:
            h = self.mem_invalid_hooks.pop(0)
            cb, = h
            self.add_mem_invalid_hook(cb)
        while self.interrupt_hooks:
            h = self.interrupt_hooks.pop(0)
            cb, ctx = h
            self.add_interrupt_hook(cb, ctx)
        while self.mem_map_hooks:
            h = self.mem_map_hooks.pop(0)
            self.add_mem_map_hook(h)

    def disasm(self, addr: int, size: int, fast=True):
        """
        Get the disassembly from an address

        args:
            addr: address to being disassebmly
            size: number of bytes to include
        return:
            A tuple of: (mnemonic, operands, and the full instruction)
        """
        try:
            return self.emu.get_disasm(addr, size, fast)
        except Exception:
            raise SpeakeasyError("Failed to disassemble at address: 0x%x" % (addr))

    def is_pe(self, data: bytes) -> bool:
        """
        Test data to see if it looks like a PE

        args:
            data: Bytes to be tested for a PE
        return:
            True is data appears to be a PE
        """
        # Check for the PE header
        if data[:2] == b'MZ':
            return True
        else:
            return False

    def load_module(self, path=None, data=None) -> PeFile:
        """
        Load a module into the speakeasy emulator

        args:
            path: Path to file to load into the emulation space
            data: Raw data to load as a module into the emulation space
        return:
            A PeFile object representing the newly loaded module
        """
        if not path and not data:
            raise SpeakeasyError('No emulation target supplied')

        if path and not os.path.exists(path):
            raise SpeakeasyError('Target file not found: %s' % (path))

        if data:
            test = data
        else:
            with open(path, 'rb') as f:
                test = f.read(4)

        self.loaded_bins.append(path)

        if not self.is_pe(test):
            raise SpeakeasyError('Target file is not a PE')

        self._init_emulator(path=path, data=data)

        return self.emu.load_module(path=path, data=data)

    @check_init
    def run_module(self, module, all_entrypoints=False, emulate_children=False) -> None:
        """
        Run a previously loaded module through the configured emulator

        args:
            module: The module whose entry point to be run
            all_entrypoints: If true, all exports will be emulated, otherwise
            just the main PE entry point is emulated.
            emulate_children: If true, any child processes created by this
            module will be emulated, otherwise, just this module is
            emulated.
        return:
            None
        """
        self._init_hooks()

        if isinstance(self.emu, Win32Emulator):
            return self.emu.run_module(module=module,
                    all_entrypoints=all_entrypoints,
                    emulate_children=emulate_children)
        else:
            return self.emu.run_module(module=module,
                    all_entrypoints=all_entrypoints)

    def load_shellcode(self, fpath, arch, data=None) -> int:
        """
        Load a shellcode blob into emulation space

        args:
            fpath: file path containing shellcode blob
            arch: Architecture (x86 | amd64) to load shellcode as
            data: bytes object containing shellcode blob
        return:
            Address of the loaded shellcode in the emulation space
        """
        self._init_emulator(is_raw_code=True)
        self.loaded_bins.append(fpath)

        return self.emu.load_shellcode(fpath, arch, data=data)

    @check_init
    def run_shellcode(self, sc_addr: int, offset=0) -> None:
        """
        Run a previously loaded shellcode blob by address

        args:
            sc_addr: address of the previously loaded shellcode blog to emulate
            offset: offset within the blob to begin emulation
        return:
            None
        """
        self._init_hooks()
        return self.emu.run_shellcode(sc_addr, offset=offset)

    @check_init
    def get_report(self) -> dict:
        """
        Get the emulation report from the emulator

        return:
            Get the raw emulation report as a python dictionary
        """
        return self.emu.get_report()

    @check_init
    def get_json_report(self) -> str:
        """
        Get the emulation report from the emulator formatted as a JSON string
        return:
            Get the emulation report as a JSON object
        """
        return self.emu.get_json_report()

    def add_api_hook(self, cb: Callable, module='', api_name='', argc=0, call_conv=None):
        """
        Set a callback to fire when a specified API is called during emulation

        args:
            cb: Callable python function to execute
            module: name of the module containing the target API
            api_name: Name of the API to hook. Wild cards (e.g. *) are supported.
            argc: force the emulator to account for this amount of arguments (for stack cleanup)
            call_conv: force the emulator to use the supplied calling convention for this hook
        return:
            Hook object for newly registered hooks
        """
        if not self.emu:
            self.api_hooks.append((cb, module, api_name, argc, call_conv))
            return
        return self.emu.add_api_hook(cb, module=module, api_name=api_name, argc=argc,
                                     call_conv=call_conv, emu=self)

    def resume(self, addr, count=-1):
        """
        Resume emulating at the specified address

        args:
            addr: Address to being emulation at
            count: number of instructions
        return:
            None
        """
        self.emu.run_complete = False
        self.emu.resume(addr, count=count)

    def stop(self) -> None:
        """
        Stops emulation
        """
        return self.emu.stop()

    def shutdown(self) -> None:
        """
        Closes the emulation instance
        """
        # TODO
        return

    def call(self, addr: int, params=[]) -> None:
        """
        Start emulating at the specified address

        args:
            addr: Address to being emulation at
            params: list of arguments to push onto the stack for the call
        return:
            None
        """
        return self.emu.call(addr, params=params)

    def add_code_hook(self, cb: Callable, begin=1, end=0, ctx={}):
        """
        Set a callback to fire for every CPU instruction that is emulated

        args:
            cb: Callable python function to execute
            begin: beginning of the address range to hook
            end: end of the address range to hook
            ctx: Optional context to pass back and forth between the hook function
        return:
            Hook object for newly registered hooks
        """
        if not self.emu:
            self.code_hooks.append((cb, begin, end, ctx))
            return
        return self.emu.add_code_hook(cb, begin=begin, end=end, ctx=ctx, emu=self)

    def add_dyn_code_hook(self, cb: Callable, ctx={}):
        """
        Set a callback to fire when dynamically generated/copied code is executed

        args:
            cb: Callable python function to execute
            ctx: Optional context to pass back and forth between the hook function
        return:
            Hook object for newly registered hooks
        """
        if not self.emu:
            self.dyn_code_hooks.append((cb, ctx))
            return
        return self.emu.add_dyn_code_hook(cb, ctx=ctx, emu=self)

    def add_mem_read_hook(self, cb: Callable, begin=1, end=0):
        """
        Set a callback to fire when a memory address is read from

        args:
            cb: Callable python function to execute
            begin: beginning of the address range to hook
            end: end of the address range to hook
        return:
            Hook object for newly registered hooks
        """
        if not self.emu:
            self.mem_read_hooks.append((cb, begin, end))
            return
        return self.emu.add_mem_read_hook(cb, begin=begin, end=end, emu=self)

    def add_mem_write_hook(self, cb: Callable, begin=1, end=0):
        """
        Set a callback to fire when a memory address is written to

        args:
            cb: Callable python function to execute
            begin: beginning of the address range to hook
            end: end of the address range to hook
        return:
            Hook object for newly registered hooks
        """
        if not self.emu:
            self.mem_write_hooks.append((cb, begin, end))
            return
        return self.emu.add_mem_write_hook(cb, begin=begin, end=end, emu=self)

    def add_IN_instruction_hook(self, cb: Callable, begin=1, end=0):
        """
        Set a callback to fire when an IN instruction executes

        args:
            cb: Callable python function to execute
            begin: beginning of the address range to hook
            end: end of the address range to hook
        return:
            Hook object for newly registered hooks
        """
        if not self.emu:
            self.mem_write_hooks.append((cb, begin, end))
            return
        return self.emu.add_instruction_hook(cb, begin=begin, end=end, emu=self, insn=218)

    def add_SYSCALL_instruction_hook(self, cb: Callable, begin=1, end=0):
        """
        Set a callback to fire when a SYSCALL / SYSENTER instruction executes

        args:
            cb: Callable python function to execute
            begin: beginning of the address range to hook
            end: end of the address range to hook
        return:
            Hook object for newly registered hooks
        """
        if not self.emu:
            self.mem_write_hooks.append((cb, begin, end))
            return
        return self.emu.add_instruction_hook(cb, begin=begin, end=end, emu=self, insn=700)

    def add_invalid_instruction_hook(self, cb: Callable, ctx=[]):
        """
        Set a callback to fire when an invalid instruction is attempted
        to be executed

        args:
            cb: Callable python function to execute
        return:
            Hook object for newly registered hooks
        """
        if not self.emu:
            self.invalid_insn_hooks.append((cb, ctx))
            return
        return self.emu.add_invalid_instruction_hook(cb, ctx)

    def add_mem_invalid_hook(self, cb: Callable):
        """
        Get a callback for when a memory access violation occurs

        args:
            cb: Callable python function to execute
        return:
            Hook object for newly registered hooks
        """
        if not self.emu:
            self.mem_invalid_hooks.append((cb, ))
            return
        return self.emu.add_mem_invalid_hook(cb, emu=self)

    def add_interrupt_hook(self, cb: Callable, ctx={}):
        """
        Get a callback for software interrupts

        args:
            cb: Callable python function to execute
            ctx: Optional context to pass back and forth between the hook function
        return:
            Hook object for newly registered hooks
        """
        if not self.emu:
            self.interrupt_hooks.append((cb, ))
            return
        return self.emu.add_interrupt_hook(cb, ctx=ctx, emu=self)

    def get_registry_key(self, handle=0, path=''):
        """
        Get registry key by path or handle

        args:
            handle: handle corresponding for a registry key
            path: Path to a registry key
        return:
            If valid, a registry key object
        """
        return self.emu.reg_get_key(handle=handle, path=path)

    def get_address_map(self, addr: int):
        """
        Get the address mapping object associated with the specified address

        args:
            addr: Address in the emulation space
        return:
            A memory map object that holds the specified address
        """
        return self.emu.get_address_map(addr)

    def get_user_modules(self) -> list:
        """
        Get the address ranges of loaded user modules

        return:
            List of all currently loaded user modules
        """
        return self.emu.get_user_modules()

    def get_sys_modules(self) -> list:
        """
        Get the address ranges of loaded system modules

        return:
            List of all currently loaded system modules
        """
        return self.emu.get_sys_modules()

    def mem_alloc(self, size, base=None, tag='speakeasy.None') -> int:
        """
        Allocate a block of memory in the emulation space

        args:
            size: Size of requested memory block
            base: Optionally request a base address. If in use, the next nearest
                  address will be returned
            tag: Tag to assign the new memory mapping
        return:
            Address of the newly allocated memory block
        """
        return self.emu.mem_map(size, base=base, tag=tag)

    def mem_free(self, base: int) -> None:
        """
        Free a block of memory in the emulation space
        args:
            base: Address to free
        return:
            None
        """
        return self.emu.mem_free(base)

    def mem_read(self, addr: int, size: int) -> bytes:
        """
        Read bytes from a memory address

        args:
            addr: address to read bytes from
            size: number of bytes to read
        return:
            Python bytes object contained the data read
        """
        try:
            return self.emu.mem_read(addr, size)
        except Exception:
            raise SpeakeasyError("Failed to read %d bytes at address: 0x%x" % (size, addr))

    def mem_write(self, addr: int, data: bytes) -> None:
        """
        Write bytes to a memory address

        args:
            addr: address to write bytes to
            data: data to write
        return:
            None
        """
        try:
            return self.emu.mem_write(addr, data)
        except Exception:
            raise SpeakeasyError("Failed to write %d bytes at address: 0x%x" % (len(data), addr))

    def mem_cast(self, obj, addr: int):
        """
        Cast an address as an object for easy access

        args:
            obj: object to cast into
            addr: address containing the data to cast into type "obj"
        return:
            Python object based on the data located at addr
        """
        return self.emu.mem_cast(obj, addr)

    def reg_read(self, reg: str) -> int:
        """
        Read value from a register

        args:
            reg: name of the register to read from
        return:
            value contained in the requested register
        """
        return self.emu.reg_read(reg)

    def get_dyn_imports(self) -> list:
        """
        Returns the imports dynamically resolved at runtime

        return:
            List of functions that were resolved at runtime (e.g. GetProcAddress,
                                                                  MmGetSystemRoutineAddress)
        """
        return self.emu.get_dyn_imports()

    def reg_write(self, reg: str, val: int) -> None:
        """
        Write value to a register

        args:
            reg: name of the register to write to
        return:
            None
        """
        return self.emu.reg_write(reg, val)

    def get_dropped_files(self) -> list:
        """
        Get files that were written to disk during emulation

        return:
            Returns a list of files that were written by the sample
        """
        return self.emu.get_dropped_files()

    def create_file_archive(self) -> bytes:
        """
        Creates a file archive package.
        The archive contains a manifest that can be used to match dropped files
        metadata with the acquired files.

        return:
            A Bytes object containing a zip archive of dropped files
        """
        manifest = []
        _zip = BytesIO()
        files = self.get_dropped_files()

        if not files:
            return b''

        with zipfile.ZipFile(_zip, mode="w", compression=zipfile.ZIP_DEFLATED) as zf:

            for f in files:

                path = f.get_path()
                file_name = ntpath.basename(path)
                manifest.append({'path': path,
                                 'file_name': file_name,
                                 'size': f.get_size(),
                                 'sha256': f.get_hash()})
                zf.writestr(file_name, f.get_data())

            manifest = json.dumps(manifest, indent=4, sort_keys=False)
            zf.writestr('speakeasy_manifest.json', manifest)

        return _zip.getvalue()

    def get_mem_maps(self) -> list:
        """
        Get all memory maps in the emulation space

        return:
            A list of all valid memory maps from the emulator
        """
        return self.emu.get_mem_maps()

    def get_memory_dumps(self) -> tuple:
        """
        Returns all memory contents along with context information

        return:
            A generator of tuples of all valid memory with context
        """
        for mm in self.emu.get_mem_maps():
            base = mm.get_base()
            size = mm.get_size()
            tag = mm.get_tag()
            proc = mm.get_process()
            is_free = mm.is_free()
            try:
                data = self.emu.mem_read(base, size)
            except Exception:
                continue
            yield (tag, base, size, is_free, proc, data)

    def read_mem_string(self, address: int, width=1, max_chars=0) -> str:
        """
        Read a string from emulated memory

        args:
            address: address of the string to read
            width: character width
            max_chars: maximum characters to read, 0 reads until null terminator

        return:
            decoded string
        """
        return self.emu.read_mem_string(address, width, max_chars)

    def get_symbols(self) -> dict:
        """
        Returns a dictionary of symbol information

        return:
            a dictionary of symbol information
        """
        return self.emu.symbols

    def get_ret_address(self) -> int:
        """
        Returns the value stored at the top of the stack

        return:
            value stored at the top of the stack
        """
        return self.emu.get_ret_address()

    def push_stack(self, val: int) -> None:
        """
        Put a value on the stack and adjust the stack pointer

        args:
            val: value to push to the stack
        return:
            None
        """
        self.emu.push_stack(val)

    def pop_stack(self) -> int:
        """
        Get value from the stack and adjust the stack pointer

        return:
            value stored at the top of the stack
        """
        return self.emu.pop_stack()

    def get_stack_ptr(self) -> int:
        """
        Get the current address of the stack pointer

        return:
            address of stack pointer
        """
        return self.emu.get_stack_ptr()

    def set_stack_ptr(self, addr: int) -> None:
        """
        Set the current address of the stack pointer

        args:
            addr: address to set the stack pointer to
        return:
            None
        """
        self.emu.set_stack_ptr(addr)

    def set_pc(self, addr: int) -> None:
        """
        Set the value of the current program counter

        args:
            addr: address to set the program counter to
        return:
            None
        """
        self.emu.set_pc(addr)

    def reset_stack(self, base: int) -> tuple:
        """
        Reset stack to the supplied base address

        args:
            base: stack base address
        return:
            base, ptr
        """
        return self.emu.reset_stack(base)

    def get_stack_base(self) -> int:
        """
        Get the base address of the stack

        return:
            base address of stack
        """
        return self.emu.stack_base

    def get_arch(self) -> int:
        """
        Get the architecture of the emulator

        return:
            emulator architecture constant value
        """
        return self.emu.get_arch()

    def get_ptr_size(self) -> int:
        """
        Get the size of a pointer

        return:
            pointer size
        """
        return self.emu.ptr_size

    def get_all_registers(self) -> dict:
        """
        Get the state of all registers

        return:
            Dict containing emulation register states
        """
        return self.emu.get_register_state()

    def get_symbol_from_address(self, address: int) -> str:
        """
        If the supplied address is related to a known symbol, look it up here

        args:
            address: address to lookup

        return:
            symbol name
        """
        return self.emu.get_symbol_from_address(address)

    def is_address_valid(self, address: int) -> bool:
        """
        Was this address previously reserved or mapped?

        args:
            address: address to check

        return:
            True if address is valid, false otherwise
        """
        return self.emu.is_address_valid(address)

    def add_mem_map_hook(self, cb: Callable, begin=1, end=0):
        """
        Set a callback to fire when a memory address is mapped

        args:
            cb: Callable python function to execute
            begin: beginning of the address range to hook
            end: end of the address range to hook
        return:
            Hook object for newly registered hooks
        """
        if not self.emu:
            self.mem_map_hooks.append((cb, begin, end))
            return
        return self.emu.add_mem_map_hook(cb, begin=begin, end=end, emu=self)

    def create_memdump_archive(self) -> bytes:
        """
        Creates a memory dump archive package of the emulated sample.
        The archive contains a manifest that can be used to match memory chunk
        metadata with the dumped binary memory files.

        return:
            Bytes object containing a zip of all memory
        """
        manifest = []
        _zip = BytesIO()

        loaded_bins = [os.path.splitext(os.path.basename(b))[0] for b in self.loaded_bins]

        with zipfile.ZipFile(_zip, mode="w", compression=zipfile.ZIP_DEFLATED) as zf:
            procs = []
            [procs.append(block[4]) for block in self.get_memory_dumps()
             if block[4] not in procs]

            for process in procs:
                memory_blocks = []
                arch = self.emu.get_arch()
                if arch == _arch.ARCH_X86:
                    arch = 'x86'
                else:
                    arch = 'amd64'

                if process:
                    pid = process.get_pid()
                    path = process.get_process_path()
                else:
                    continue

                manifest.append({'pid': pid, 'process_name': path, 'arch': arch,
                                 'memory_blocks': memory_blocks})
                for block in self.get_memory_dumps():

                    tag, base, size, is_free, _proc, data = block

                    if not tag:
                        continue
                    if _proc != process:
                        continue
                    # Ignore emulator noise such as structures created by the emulator, or
                    # modules that were loaded
                    if tag and tag.startswith('emu') and not tag.startswith('emu.shellcode.'):
                        bns = [b for b in loaded_bins if b in tag]
                        if not len(bns):
                            continue

                    h = hashlib.sha256()
                    h.update(data)
                    _hash = h.hexdigest()

                    file_name = '%s.mem' % (tag)

                    memory_blocks.append({'tag':  tag, 'base': hex(base), 'size': hex(size),
                                          'is_free': is_free, 'sha256': _hash,
                                          'file_name': file_name})
                    zf.writestr(file_name, data)

            manifest = json.dumps(manifest, indent=4, sort_keys=False)
            zf.writestr('speakeasy_manifest.json', manifest)

        return _zip.getvalue()


def validate_config(config) -> None:
    """
    Validates the given configuration objects against the built-in schemas.

    Raises jsonschema.exceptions.ValidationError on invalid configuration.
    Expose the underlying jsonschema exception due to it having lots of information
    about failures.

    On success, returns without exception.
    """
    schema_path = os.path.join(os.path.dirname(speakeasy.__file__), 'config_schema.json')
    with open(schema_path, 'r') as ff:
        schema = json.load(ff)
    validator = jsonschema.Draft7Validator(schema)
    validator.validate(config)



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import os
import ntpath
import hashlib
from collections import namedtuple

import pefile

import speakeasy.winenv.arch as _arch
import speakeasy.winenv.defs.nt.ddk as ddk
from speakeasy.struct import Enum


# GDT Constants needed to set our emulator into protected mode
# Access bits
GDT_ACCESS_BITS = Enum()
GDT_ACCESS_BITS.ProtMode32 = 0x4
GDT_ACCESS_BITS.PresentBit = 0x80
GDT_ACCESS_BITS.Ring3 = 0x60
GDT_ACCESS_BITS.Ring0 = 0
GDT_ACCESS_BITS.DataWritable = 0x2
GDT_ACCESS_BITS.CodeReadable = 0x2
GDT_ACCESS_BITS.DirectionConformingBit = 0x4
GDT_ACCESS_BITS.Code = 0x18
GDT_ACCESS_BITS.Data = 0x10

GDT_FLAGS = Enum()
GDT_FLAGS.Ring3 = 0x3
GDT_FLAGS.Ring0 = 0

IMPORT_HOOK_ADDR = 0xFEEDFACE
DEFAULT_LOAD_ADDR = 0x40000

PAGE_SIZE = 0x1000

EMU_RESERVED = 0xfeedf000
EMU_RESERVE_SIZE = 0x4000
DYM_IMP_RESERVE = EMU_RESERVED + 0x1000
EMU_CALLBACK_RESERVE = DYM_IMP_RESERVE + 0x1000
EMU_SYSCALL_RESERVE = EMU_CALLBACK_RESERVE + 0x1000

EMU_RESERVED_END = (EMU_RESERVED + EMU_RESERVE_SIZE)
EMU_RETURN_ADDR = EMU_RESERVED
EXIT_RETURN_ADDR = EMU_RETURN_ADDR + 1
SEH_RETURN_ADDR = EMU_RETURN_ADDR + 4
API_CALLBACK_HANDLER_ADDR = EMU_RETURN_ADDR + 8
IMPORT_HOOK_ADDR = EMU_RETURN_ADDR + 12

# Common blank DOS header
DOS_HEADER = b'MZ\x90\x00\x03\x00\x00\x00\x04\x00\x00\x00\xff\xff\x00\x00\xb8\x00\x00\x00' \
             b'\x00\x00\x00\x00@' + (b'\x00' * 35) + \
             b'\xb0\x00\x00\x00\x0e\x1f\xba\x0e\x00\xb4\t\xcd!\xb8\x01L\xcd!This program ' \
             b'cannot be run in DOS mode.\r\r\n$' + (b'\x00' * 39) + \
             b'Rich\xbeL\x1c\x41\x00\x00\x00\x00\x00\x00\x00\x00'

# Blank header used for a 32-bit PE header
EMPTY_PE_32 = DOS_HEADER + b'PE\x00\x00L\x01\x00\x00ABCD\x00\x00\x00\x00\x00\x00\x00\x00'   \
                           b'\xe0\x00\x03\x01\x0b\x01\x08\x00\x04\x00\x00\x00\x00\x00\x00'  \
                           b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\xd0\x01\x00\x00\xd4\x01'  \
                           b'\x00\x00\x00\x00@\x00\x01\x00\x00\x00\x01\x00\x00\x00\x04\x00' \
                           b'\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\xd4'  \
                           b'\x01\x00\x00\xd0\x01\x00\x00\x00\x00\x00\x00\x02\x00\x00\x04'  \
                           b'\x00\x00\x10\x00\x00\x10\x00\x00\x00\x00\x10\x00\x00\x10\x00'  \
                           b'\x00\x00\x00\x00\x00\x10' + (b'\x00' * 131)

# Blank header used for a 64-bit PE header
EMPTY_PE_64 = DOS_HEADER + b'PE\x00\x00d\x86\x00\x00ABCD\x00\x00\x00\x00\x00\x00\x00\x00'   \
                           b'\xf0\x00\x03\x10\x0b\x02\x08\x00\x04\x00\x00\x00\x00\x00\x00'  \
                          b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00@' \
                           b'\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x06\x00'  \
                           b'\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\xb8'  \
                           b'\x01\x00\x00\x00\x00\x00\x00AAAA\x02\x00\x00\x04\x00\x00\x10'  \
                           b'\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00'  \
                           b'\x10\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00'  \
                           b'\x00\x00\x00\x10' + (b'\x00' * 131)


def normalize_dll_name(name):
    ret = name

    # Funnel CRTs into a single handler
    if name.lower().startswith(('api-ms-win-crt', 'vcruntime', 'ucrtbased', 'ucrtbase', 'msvcr', 'msvcp')):
        ret = 'msvcrt'

    # Redirect windows sockets 1.0 to windows sockets 2.0
    elif name.lower().startswith(('winsock', 'wsock32')):
        ret = 'ws2_32'

    elif name.lower().startswith('api-ms-win-core'):
        ret = 'kernel32'

    return ret


class PeParseException(Exception):
    pass


class PeFile(pefile.PE):
    """
    Class the will represent all PE files loaded into the emulator
    """
    def __init__(self, path=None, data=None, imp_id=IMPORT_HOOK_ADDR,
                 imp_step=4, emu_path='', fast_load=False):

        super(PeFile, self).__init__(name=path, data=data, fast_load=fast_load)

        if 0 == self.OPTIONAL_HEADER.ImageBase:
            self.relocate_image(DEFAULT_LOAD_ADDR)
            super(PeFile, self).__init__(name=None, data=self.write())

        self.imp_id = imp_id
        self.imp_step = imp_step
        self.file_size = 0
        self.base = self.OPTIONAL_HEADER.ImageBase
        self.hash = self._hash_pe(path=path, data=data)
        self.imports = self._get_pe_imports()
        self.exports = self._get_pe_exports()
        self.mapped_image = self.get_memory_mapped_image(max_virtual_address=0xf0000000)
        # self.mapped_image = None
        self.image_size = self.OPTIONAL_HEADER.SizeOfImage
        self.import_table = {}
        self.is_mapped = True
        self.pe_sections = self._get_pe_sections()
        self.ep = self.OPTIONAL_HEADER.AddressOfEntryPoint
        self.stack_commit = self.OPTIONAL_HEADER.SizeOfStackCommit
        self.path = ''
        self.name = ''
        if path:
            self.path = os.path.abspath(path)
        self.emu_path = emu_path
        self.arch = self._get_architecture()
        if self.arch == _arch.ARCH_X86:
            self.ptr_size = 4
        else:
            self.ptr_size = 8

        self._patch_imports()

    def get_tls_callbacks(self):
        """
        Get the TLS callbacks for a PE (if any)
        """
        max_tls_callbacks = 100
        callbacks = []
        if hasattr(self, 'DIRECTORY_ENTRY_TLS'):
            rva = (self.DIRECTORY_ENTRY_TLS.struct.AddressOfCallBacks -
                   self.OPTIONAL_HEADER.ImageBase)

            for i in range(max_tls_callbacks):
                ptr = self.get_data(rva + self.ptr_size * i, self.ptr_size)
                ptr = int.from_bytes(ptr, 'little')
                if ptr == 0:
                    break
                callbacks.append(ptr)
        return callbacks

    def get_emu_path(self):
        """
        Get the path of the module (as it appears to the emulated binary)
        """
        return self.emu_path

    def set_emu_path(self, path):
        self.emu_path = path

    def _hash_pe(self, path=None, data=None):
        hasher = hashlib.sha256()
        buf = b''
        if path:
            with open(path, 'rb') as f:
                buf = f.read()
        elif data:
            buf = data

        hasher.update(buf)
        self.file_size = len(buf)
        return hasher.hexdigest()

    def _get_pe_imports(self):
        pe = self
        imports = {}

        try:
            pe.DIRECTORY_ENTRY_IMPORT
        except Exception:
            return imports

        for entry in pe.DIRECTORY_ENTRY_IMPORT:
            dll = entry.dll
            dll = dll.decode('utf-8')
            dll = os.path.splitext(dll)[0]
            for imp in entry.imports:
                if imp.import_by_ordinal:
                    func_name = 'ordinal_%d' % (imp.ordinal)
                    imports.update({imp.address: (dll, func_name)})
                else:
                    func_name = imp.name.decode('utf-8')
                    imports.update({imp.address: (dll, func_name)})
        return imports

    def get_exports(self):
        self.exports = self._get_pe_exports()

        return self.exports

    def _get_pe_exports(self):
        pe = self
        exports = []
        try:
            pe.DIRECTORY_ENTRY_EXPORT
        except Exception:
            return exports

        for exp in pe.DIRECTORY_ENTRY_EXPORT.symbols:
            entry = namedtuple('export', ['name', 'address', 'forwarder', 'ordinal'])
            entry.name = exp.name
            entry.address = exp.address + pe.get_base()
            entry.forwarder = exp.forwarder
            entry.ordinal = exp.ordinal
            if entry.name:
                entry.name = entry.name.decode('utf-8')
            exports.append(entry)
        return exports

    def _get_pe_sections(self):
        pe = self
        sections = []
        for section in pe.sections:
            sect = (section.Name, section.VirtualAddress,
                    section.Misc_VirtualSize, section.SizeOfRawData)
            sections.append(sect)
        return sections

    def get_sections(self):
        return self.sections

    def get_section_by_name(self, name):
        sect = [s for s in self.get_sections() if s.Name.decode('utf-8').strip('\x00') == name]
        if sect:
            return sect[0]

    def _get_architecture(self):
        # 0x010b: PE32, 0x020b: PE32+ (64 bit)
        magic = self.OPTIONAL_HEADER.Magic
        if magic & ddk.PE32_BIT:
            return _arch.ARCH_X86
        elif magic & ddk.PE32_PLUS_BIT:
            return _arch.ARCH_AMD64
        else:
            raise ValueError('Unsupported architecture: 0x%x' % (magic))

    def _patch_imports(self):
        """
        Imports are patched with invalid memory addresses. When the API is called
        by the emulated binary, the invalid memory fetch callback will trigger,
        allowing us to handle the Windows API within the emulator
        """
        if not self.imports:
            return

        if not self.mapped_image:
            raise ValueError('PE image has not been mapped yet')

        for addr, imp in self.imports.items():
            tmp = bytearray(self.mapped_image)
            offset = addr - self.base
            tmp[offset: offset + self.ptr_size] = \
                self.imp_id.to_bytes(self.ptr_size, 'little')
            self.mapped_image = bytes(tmp)

            self.import_table.update({self.imp_id: imp})
            self.imp_id += self.imp_step

    def get_export_by_name(self, name):
        for exp in self.get_exports():
            if name == exp.name:
                return exp.address

    def get_raw_data(self):
        return self.get_memory_mapped_image()

    def find_bytes(self, pattern, offset=0):
        return self.get_raw_data().find(pattern, offset)

    def set_bytes(self, offset, pattern):
        self.set_bytes_at_offset(offset, pattern)

    def get_ptr_size(self):
        return self.ptr_size

    def get_base(self):
        return self.base

    def get_base_name(self):
        fn = os.path.basename(self.path)
        bn = os.path.splitext(fn)[0]
        return bn

    def get_image_size(self):
        return self.image_size

    def is_decoy(self):
        return False

    def is_driver(self):
        rv = super(PeFile, self).is_driver()
        if rv:
            return rv

        system_DLLs = set((b'ntoskrnl.exe', b'hal.dll', b'ndis.sys',
                           b'bootvid.dll', b'kdcom.dll', b'win32k.sys'))

        if hasattr(self, 'DIRECTORY_ENTRY_IMPORT'):
            if system_DLLs.intersection(
                    [imp.dll.lower() for imp in self.DIRECTORY_ENTRY_IMPORT]):
                return True

        if self.OPTIONAL_HEADER.Subsystem == pefile.SUBSYSTEM_TYPE['IMAGE_SUBSYSTEM_NATIVE'] \
           and self.ep == 0:
            return True

    def is_dotnet(self):
        """
        Is the current PE file a .NET assembly?
        """
        for addr, imp in self.imports.items():
            dll, func = imp
            if dll == 'mscoree' and func in ['_CorExeMain', '_CorDllMain']:
                return True
        return False

    def has_reloc_table(self):
        return len(self.OPTIONAL_HEADER.DATA_DIRECTORY) >= 6 and \
                self.OPTIONAL_HEADER.DATA_DIRECTORY[5].Size > 0

    def rebase(self, to):
        self.relocate_image(to)

        self.base = to
        self.ep = self.OPTIONAL_HEADER.AddressOfEntryPoint

        # After relocation, generate a new memory mapped image
        self.mapped_image = self.get_memory_mapped_image(max_virtual_address=0xf0000000)

        self.pe_sections = self._get_pe_sections()
        self.imports = self._get_pe_imports()
        self.exports = self._get_pe_exports()
        self._patch_imports()

        return

class DecoyModule(PeFile):
    """
    Class that represents "decoy" modules that are loaded into emulated memory.
    We use decoy modules so that shellcode
    (or other modules) can parse the PE file to resolve exports.
    """
    def __init__(self, path=None, data=None, fast_load=True, base=0, emu_path='', is_jitted=False):
        self.image_size = 0
        self.ep = 0
        self.is_jitted = is_jitted
        if path or data:
            super(DecoyModule, self).__init__(path=path, data=data, fast_load=fast_load)

        if data:
            self.image_size = len(data)

        self.decoy_base = base
        self.decoy_path = emu_path
        self.base_name = ''
        self.is_mapped = False
        self.data = b''

    def get_memory_mapped_image(self, max_virtual_address=0x10000000, base=None):
        mmi = super(DecoyModule, self).get_memory_mapped_image(max_virtual_address, base)
        if self.is_jitted and len(mmi) < len(self.__data__):
            return self.__data__
        return mmi

    def get_base(self):
        return self.decoy_base

    def get_emu_path(self):
        return self.decoy_path

    def get_base_name(self):
        p = self.get_emu_path()
        img = ntpath.basename(p)
        bn = os.path.splitext(img)[0]
        return bn

    def get_ep(self):
        return self.get_base() + self.ep

    def is_decoy(self):
        return True


class JitPeFile(object):
    '''
    Class used to rapidly assemble a decoy PE that will only contain an export table
    so malware can parse it.
    '''
    def __init__(self, arch):

        if arch == _arch.ARCH_X86:
            self.pattern_size = 9
            husk = EMPTY_PE_32
        else:
            self.pattern_size = 12
            husk = EMPTY_PE_64

        self.arch = arch

        self.basepe = pefile.PE(data=husk, fast_load=True)

    def get_section_by_name(self, pe, name):
        '''
        Get a PE section by name
        '''
        for sect in pe.sections:
            if sect.Name.decode('utf-8').strip('\x00') == name.strip('\x00'):
                return self.cast_section(sect.get_file_offset())

    def get_raw_pe(self):
        '''
        Get the raw data associated with a decoy PE
        '''
        return self.basepe.__data__

    def update(self):
        '''
        Update the raw data associated with a decoy PE
        '''
        self.basepe = pefile.PE(None, self.basepe.write(), fast_load=True)
        self.update_image_size()

    def cast_section(self, offset=None):
        '''
        Get a section from a given offset
        '''
        if offset is None:
            offset = self.get_current_offset()
            data = pefile.Structure(self.basepe.__IMAGE_SECTION_HEADER_format__).sizeof() * b'\x00'
        else:
            data = self.basepe.get_data(offset, pefile.Structure(self.basepe.__IMAGE_SECTION_HEADER_format__).sizeof()) # noqa

        sect = self.basepe.__unpack_data__(self.basepe.__IMAGE_SECTION_HEADER_format__, data,
                                           offset)
        return sect

    def update_image_size(self):
        '''
        Update the size of the image within the optional header
        '''
        self.basepe.OPTIONAL_HEADER.SizeOfImage = (len(self.basepe.get_memory_mapped_image()) +
                                                   PAGE_SIZE)

    def add_section(self, name, chars=0x40000040):
        '''
        Add a section to the decoy PE
        '''
        new_sect = self.cast_section()
        new_sect.Name = name.encode('utf-8')
        new_sect.Characteristics = chars

        hdr_size = pefile.Structure(self.basepe.__IMAGE_SECTION_HEADER_format__).sizeof()

        self.basepe.OPTIONAL_HEADER.SizeOfHeaders += hdr_size
        self.basepe.OPTIONAL_HEADER.SizeOfImage += hdr_size
        self.basepe.FILE_HEADER.NumberOfSections += 1

        self.update()
        return new_sect

    def get_current_offset(self):
        '''
        Get the current offset (or size) of the PE
        '''
        return len(self.basepe.__data__)

    def append_data(self, data):
        '''
        Append data to the decoy PE
        '''
        self.basepe.__data__ += data

    def get_exports_size(self, name, exports):
        '''
        Get the total size of the export directory
        '''
        # Get the total size needed for the new export section
        exp_size = pefile.Structure(self.basepe.__IMAGE_EXPORT_DIRECTORY_format__).sizeof()

        exp_size += (len(name) + 1)

        for exp in exports:
            exp_size += len(exp) + 1
            exp_size += (0x4 + 0x4 + 0x4)

        return exp_size

    def get_decoy_pe_image(self, mod_name, exports):

        self.add_section(name='.text')
        self.add_section(name='.edata')

        self.init_text_section(exports)
        self.init_export_section(mod_name.encode('utf-8'), exports)
        self.update()
        return self.get_raw_pe()

    def init_export_section(self, name, exports):
        '''
        Initialize and add the export table to the PE
        '''
        exports_size = self.get_exports_size(name, exports)

        dest_exp_sect = self.get_section_by_name(self.basepe, '.edata')

        dest_exp_sect.Misc_VirtualSize = exports_size
        dest_exp_sect.Misc_PhysicalAddress = 0
        dest_exp_sect.VirtualAddress = self.get_current_offset()
        dest_exp_sect.SizeOfRawData = exports_size
        dest_exp_sect.PointerToRawData = self.get_current_offset()

        self.basepe.OPTIONAL_HEADER.SizeOfInitializedData += exports_size

        export_dir = self.basepe.OPTIONAL_HEADER.DATA_DIRECTORY[0]

        export_dir.VirtualAddress = dest_exp_sect.VirtualAddress
        export_dir.Size = exports_size

        offset = self.get_current_offset()
        self.append_data(b'\x00' * exports_size)

        dest_export_dir = self.basepe.__unpack_data__(self.basepe.__IMAGE_EXPORT_DIRECTORY_format__, # noqa
                                                      pefile.Structure(self.basepe.__IMAGE_EXPORT_DIRECTORY_format__).sizeof() * b'\x00', # noqa
                                                      offset)
        offset += pefile.Structure(self.basepe.__IMAGE_EXPORT_DIRECTORY_format__).sizeof()

        dest_export_dir.Characteristics = 0
        dest_export_dir.TimeDateStamp = 0xD1234567
        dest_export_dir.MajorVersion = 0
        dest_export_dir.MinorVersion = 0
        dest_export_dir.Base = 1
        dest_export_dir.NumberOfFunctions = len(exports)
        dest_export_dir.NumberOfNames = len(exports)

        # Set the address of functions array
        num_funcs = dest_export_dir.NumberOfFunctions
        funcs_offset = offset
        names_offset = funcs_offset + (4 * num_funcs)
        ord_offset = names_offset + (4 * num_funcs)
        strings_offset = ord_offset + (2 * num_funcs)

        dest_export_dir.Name = strings_offset

        dest_export_dir.AddressOfFunctions = offset
        dest_export_dir.AddressOfNames = names_offset
        dest_export_dir.AddressOfNameOrdinals = ord_offset

        # Set the export name
        self.basepe.set_bytes_at_offset(strings_offset, name)
        strings_offset += len(name) + 1

        ep = self.basepe.OPTIONAL_HEADER.AddressOfEntryPoint

        for i, exp in enumerate(exports):

            exp = exp.encode('utf-8')

            # Add fluff to pass forwarded export checks
            self.append_data(b'\x00' * len(exports))

            # Add the function addresses
            self.basepe.set_dword_at_offset(funcs_offset, ep)
            funcs_offset += 4
            if funcs_offset > self.get_current_offset():
                raise Exception('Functions offset exceeds total PE size')

            # Add the ordinals
            self.basepe.set_word_at_offset(ord_offset, (i + 1) - dest_export_dir.Base)
            ord_offset += 2
            if ord_offset > self.get_current_offset():
                raise Exception('Ordinals offset exceeds total PE size')

            # Add the function names in
            if strings_offset > self.get_current_offset():
                raise Exception('Export string offset exceeds total PE size')
            self.basepe.set_dword_at_offset(names_offset, strings_offset)
            names_offset += 4
            self.basepe.set_bytes_at_offset(strings_offset, exp)
            strings_offset += len(exp) + 1

            ep += self.pattern_size

        if strings_offset:
            self.basepe.__data__ = self.basepe.__data__[:strings_offset]
        self.update()

    def init_text_section(self, names):
        '''
        Initialize and add the text section to the PE
        '''
        pattern = b''
        # Add placeholder code in case emulated samples want to hook the function
        if self.arch == _arch.ARCH_X86:
            for i in range(len(names)):
                pattern += (b'\x89\xff\x90\xB8' + i.to_bytes(4, 'little') + b'\xc3')
        else:
            for i in range(len(names)):
                pattern += (b'\x48\x89\xFF\x90\x48\xC7\xC0' + i.to_bytes(4, 'little') + b'\xc3')

        if pattern:
            sect = self.get_section_by_name(self.basepe, '.text')
            sect.VirtualAddress = self.get_current_offset()
            sect.Misc_VirtualSize = len(pattern)
            sect.Misc_PhysicalAddress = 0
            sect.SizeOfRawData = len(pattern)
            sect.PointerToRawData = self.get_current_offset()
            self.basepe.OPTIONAL_HEADER.AddressOfEntryPoint = sect.VirtualAddress
            self.append_data(pattern)
        self.update()



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import EmuStruct, EmuUnion, Ptr
import ctypes as ct

from speakeasy.winenv.defs import * # noqa


class KSYSTEM_TIME(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.LowPart = ct.c_uint32
        self.High1Time = ct.c_uint32
        self.High2Time = ct.c_uint32


class SSDT(EmuStruct):  # KeServiceDescriptorTable
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.pServiceTable = Ptr
        self.pCounterTable = Ptr
        self.NumberOfServices = ct.c_uint32
        self.pArgumentTable = Ptr


class UNICODE_STRING(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Length = ct.c_uint16
        self.MaximumLength = ct.c_uint16
        self.Buffer = Ptr


class DeviceIoControl(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.OutputBufferLength = ct.c_uint32
        self.InputBufferLength = ct.c_uint32
        self.IoControlCode = ct.c_uint32
        self.Type3InputBuffer = Ptr


class STRING(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Length = ct.c_uint16
        self.MaximumLength = ct.c_uint16
        self.Buffer = Ptr


class SYSTEM_MODULE(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Reserved = Ptr * 2
        self.Base = Ptr
        self.Size = ct.c_uint32
        self.Flags = ct.c_uint32
        self.Index = ct.c_uint16
        self.Unknown = ct.c_uint16
        self.LoadCount = ct.c_uint16
        self.ModuleNameOffset = ct.c_uint16
        self.ImageName = ct.c_uint8 * 256


class SYSTEM_TIMEOFDAY_INFORMATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.BootTime = ct.c_uint64
        self.CurrentTime = ct.c_uint64
        self.TimeZoneBias = ct.c_uint64
        self.TimeZoneId = ct.c_uint32
        self.Reserved = ct.c_uint32
        self.BootTimeBias = ct.c_uint64
        self.SleepTimeBias = ct.c_uint64


class SYSTEM_PROCESS_INFORMATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.NextEntryOffset = ct.c_uint32
        self.NumberOfThreads = ct.c_uint32
        self.Reserved1 = ct.c_uint8 * 48
        self.ImageName = UNICODE_STRING
        self.BasePriority = ct.c_uint32
        self.UniqueProcessId = Ptr
        self.InheritedFromUniqueProcessId = Ptr
        self.HandleCount = ct.c_uint32
        self.SessionId = ct.c_uint32
        self.UniqueProcessKey = Ptr
        self.PeakVirtualSize = Ptr
        self.VirtualSize = Ptr
        self.PageFaultCount = ct.c_uint32

        self.PeakWorkingSetSize = Ptr
        self.WorkingSetSize = Ptr
        self.QuotaPeakPagedPoolUsage = Ptr
        self.QuotaPagedPoolUsage = Ptr
        self.QuotaPeakNonPagedPoolUsage = Ptr
        self.QuotaNonPagedPoolUsage = Ptr
        self.PagefileUsage = Ptr

        self.PeakPagefileUsage = Ptr
        self.PrivatePageCount = Ptr
        self.Reserved7 = LARGE_INTEGER * 6


class SYSTEM_THREAD_INFORMATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Reserved1 = LARGE_INTEGER * 3
        self.Reserved2 = ct.c_uint32
        self.StartAddress = Ptr
        self.ClientId = CLIENT_ID
        self.Priority = ct.c_uint32
        self.BasePriority = ct.c_uint32
        self.ContextSwitches = ct.c_uint32
        self.ThreadState = ct.c_uint32
        self.WaitReason = ct.c_uint32


class CLIENT_ID(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.UniqueProcess = ct.c_uint32
        self.UniqueThread = ct.c_uint32


class MDL(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Next = Ptr
        self.Size = ct.c_uint16
        self.MdlFlags = ct.c_uint16
        self.Process = Ptr
        self.MappedSystemVa = Ptr
        self.StartVa = Ptr
        self.ByteCount = ct.c_uint32
        self.ByteOffset = ct.c_uint32


class KIDTENTRY(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.OffsetLow = ct.c_uint16
        self.Selector = ct.c_uint16
        self.Base = ct.c_uint32


class KIDTENTRY64(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.OffsetLow = ct.c_uint16
        self.Selector = ct.c_uint16
        self.Reserved0 = ct.c_uint16
        self.OffsetMiddle = ct.c_uint16
        self.OffsetHigh = ct.c_uint32
        self.Reserved1 = ct.c_uint32


class ETHREAD(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Data = ct.c_uint8 * 4096


class EPROCESS(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Data = ct.c_uint8 * 4096


class KEVENT(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Data = ct.c_uint8 * 4096


class MUTANT(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Data = ct.c_uint8 * 4096


class RTL_OSVERSIONINFOW(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.dwOSVersionInfoSize = ct.c_uint32
        self.dwMajorVersion = ct.c_uint32
        self.dwMinorVersion = ct.c_uint32
        self.dwBuildNumber = ct.c_uint32
        self.dwPlatformId = ct.c_uint32
        self.szCSDVersion = ct.c_uint8 * 256


class RTL_OSVERSIONINFOEXW(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.dwOSVersionInfoSize = ct.c_uint32
        self.dwMajorVersion = ct.c_uint32
        self.dwMinorVersion = ct.c_uint32
        self.dwBuildNumber = ct.c_uint32
        self.dwPlatformId = ct.c_uint32
        self.szCSDVersion = ct.c_uint8 * 256
        self.wServicePackMajor = ct.c_uint16
        self.wServicePackMinor = ct.c_uint16
        self.wSuiteMask = ct.c_uint16
        self.wProductType = ct.c_uint8
        self.wReserved = ct.c_uint8


class IDT(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Limit = ct.c_uint16
        self.Descriptors = Ptr


class KAPC(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Type = ct.c_uint8
        self.SpareByte0 = ct.c_uint8
        self.Size = ct.c_uint8
        self.SpareByte1 = ct.c_uint8
        self.SpareLong0 = ct.c_uint32
        self.Thread = Ptr
        self.ApcListEntry = LIST_ENTRY
        self.KernelRoutine = Ptr
        self.RundownRoutine = Ptr
        self.NormalRoutine = Ptr
        self.NormalContext = Ptr
        self.SystemArgument1 = Ptr
        self.SystemArgument2 = Ptr
        self.ApcStateIndex = ct.c_uint8
        self.ApcMode = ct.c_uint8
        self.Inserted = ct.c_uint8


class OBJECT_ATTRIBUTES(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Length = ct.c_uint32
        self.RootDirectory = Ptr
        self.ObjectName = Ptr
        self.Attributes = ct.c_uint32
        self.SecurityDescriptor = Ptr
        self.SecurityQualityOfService = Ptr


class FILE_STANDARD_INFORMATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.AllocationSize = LARGE_INTEGER
        self.EndOfFile = LARGE_INTEGER
        self.NumberOfLinks = ct.c_uint32
        self.DeletePending = ct.c_uint8
        self.Directory = ct.c_uint8


class DESCRIPTOR_TABLE(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        if ptr_size == 4:
            self.Table = KIDTENTRY * 256
        else:
            self.Table = KIDTENTRY64 * 256


class DRIVER_OBJECT(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Type = ct.c_uint16
        self.Size = ct.c_uint16
        self.DeviceObject = Ptr
        self.Flags = ct.c_uint32
        self.DriverStart = Ptr
        self.DriverSize = ct.c_uint32
        self.DriverSection = Ptr
        self.DriverExtension = Ptr
        self.DriverName = UNICODE_STRING
        self.HardwareDatabase = Ptr
        self.FastIoDispatch = Ptr
        self.DriverInit = Ptr
        self.DriverStartIo = Ptr
        self.DriverUnload = Ptr
        self.MajorFunction = Ptr * 28


class KDEVICE_QUEUE(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Type = ct.c_uint16
        self.Size = ct.c_uint16
        self.DeviceListHead = LIST_ENTRY
        self.Lock = ct.c_uint64
        self.Busy = ct.c_uint8


class KDPC(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Type = ct.c_uint8
        self.Importance = ct.c_uint8
        self.Number = ct.c_uint16
        self.DpcListEntry = LIST_ENTRY
        self.DeferredRoutine = Ptr
        self.DeferredContext = Ptr
        self.SystemArgument1 = Ptr
        self.SystemArgument2 = Ptr
        self.DpcData = Ptr


class DEVICE_OBJECT(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Type = ct.c_uint16
        self.Size = ct.c_uint16
        self.ReferenceCount = ct.c_uint32
        self.DriverObject = Ptr
        self.NextDevice = Ptr
        self.AttachedDevice = Ptr
        self.CurrentIrp = Ptr
        self.Timer = Ptr
        self.Flags = ct.c_uint32
        self.Characteristics = ct.c_uint32
        self.Vpb = Ptr
        self.DeviceExtension = Ptr
        self.DeviceType = ct.c_uint32
        self.StackSize = ct.c_uint8
        self.Queue = LIST_ENTRY
        self.AlignmentRequirement = ct.c_uint32
        self.DeviceQueue = KDEVICE_QUEUE
        self.Dpc = KDPC
        self.ActiveThreadCount = ct.c_uint32
        self.SecurityDescriptor = Ptr
        self.DeviceLock = KEVENT
        self.SectorSize = ct.c_uint16
        self.Spare1 = ct.c_uint16
        self.DeviceObjectExtension = Ptr
        self.Reserved = Ptr


class FILE_OBJECT(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Type = ct.c_uint16
        self.Size = ct.c_uint16
        self.DeviceObject = Ptr
        self.Vpb = Ptr
        self.FsContext = Ptr
        self.FsContext2 = Ptr
        self.SectionObjectPointer = Ptr
        self.PrivateCacheMap = Ptr
        self.FinalStatus = ct.c_uint32
        self.RelatedFileObject = Ptr
        self.LockOperation = ct.c_uint8
        self.DeletePending = ct.c_uint8
        self.ReadAccess = ct.c_uint8
        self.WriteAccess = ct.c_uint8
        self.DeleteAccess = ct.c_uint8
        self.SharedRead = ct.c_uint8
        self.SharedWrite = ct.c_uint8
        self.SharedDelete = ct.c_uint8
        self.Flags = ct.c_uint32
        self.FileName = UNICODE_STRING
        self.CurrentByteOffset = LARGE_INTEGER
        self.Waiters = ct.c_uint32
        self.Busy = ct.c_uint32
        self.LastLock = Ptr
        self.Lock = KEVENT
        self.Event = KEVENT
        self.CompletionContext = Ptr
        self.IrpListLock = ct.c_uint32
        self.IrpList = LIST_ENTRY
        self.FileObjectExtension = Ptr


class IO_PARAMETERS(EmuUnion):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.DeviceIoControl = DeviceIoControl(ptr_size).get_cstruct()


class IO_STACK_LOCATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.MajorFunction = ct.c_uint8
        self.MinorFunction = ct.c_uint8
        self.Flags = ct.c_uint8
        self.Control = ct.c_uint8
        if ptr_size == 8:
            self._padding = ct.c_uint8 * 8
        self.Parameters = IO_PARAMETERS
        self.DeviceObject = Ptr
        self.FileObject = Ptr
        self.CompletionRoutine = Ptr
        self.Context = Ptr


class IRP_OVERLAY(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.UserApcRoutine = Ptr
        self.UserApcContext = Ptr


class IRP(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Type = ct.c_uint16
        self.Size = ct.c_uint16
        self.MdlAddress = Ptr
        self.Flags = ct.c_uint32
        self.AssociatedIrp = Ptr
        self.ThreadListEntry = LIST_ENTRY
        self.IoStatus = IO_STATUS_BLOCK
        self.RequestorMode = ct.c_uint8
        self.PendingReturned = ct.c_uint8
        self.StackCount = ct.c_uint8
        self.CurrentLocation = ct.c_uint8
        self.Cancel = ct.c_uint8
        self.CancelIrql = ct.c_uint8
        self.ApcEnvironment = ct.c_uint8
        self.AllocationFlags = ct.c_uint8
        self.UserIosb = Ptr
        self.UserEvent = Ptr
        self.Overlay = IRP_OVERLAY
        self.CancelRoutine = Ptr
        self.UserBuffer = Ptr
        self.Tail = IRP_TAIL


class IO_STATUS_BLOCK(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Status = Ptr
        self.Information = Ptr


class KDEVICE_QUEUE_ENTRY(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.DeviceListEntry = LIST_ENTRY
        self.SortKey = ct.c_uint32
        self.Inserted = ct.c_uint8


class TAIL_OVERLAY(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.DeviceQueueEntry = KDEVICE_QUEUE_ENTRY
        if ptr_size == 8:
            self.padding = ct.c_uint8 * 8

        self.Reserved1 = Ptr * 2
        self.ListEntry = LIST_ENTRY
        self.CurrentStackLocation = Ptr
        self.Reserved2 = Ptr


class IRP_TAIL(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Overlay = TAIL_OVERLAY


class LIST_ENTRY(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Flink = Ptr
        self.Blink = Ptr


class NT_TIB(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.ExceptionList = Ptr
        self.StackBase = Ptr
        self.StackLimit = Ptr
        self.Reserved1 = Ptr
        self.Reserved2 = Ptr
        self.Reserved3 = Ptr
        self.Self = Ptr


class TEB(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.NtTib = NT_TIB
        self.EnvironmentPointer = Ptr
        self.ClientId = CLIENT_ID
        if ptr_size == 8:
            self.pad0 = Ptr
        self.ActiveRpcHandle = Ptr
        self.ThreadLocalStoragePointer = Ptr
        self.ProcessEnvironmentBlock = Ptr
        self.LastErrorValue = ct.c_uint32
        self.CountOfOwnedCriticalSections = ct.c_uint32
        self.CsrClientThread = Ptr
        self.Win32ThreadInfo = Ptr
        self.User32Reserved = ct.c_uint32 * 26
        self.UserReserved = ct.c_uint32 * 5
        self.WOW32Reserved = Ptr
        self.CurrentLocale = ct.c_uint32


class PEB(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.InheritedAddressSpace = ct.c_uint8
        self.ReadImageFileExecOptions = ct.c_uint8
        self.BeingDebugged = ct.c_uint8
        self.BitField = ct.c_uint8
        self.Mutant = Ptr
        self.ImageBaseAddress = Ptr
        self.Ldr = Ptr
        self.ProcessParameters = Ptr
        self.SubSystemData = Ptr
        self.ProcessHeap = Ptr
        self.FastPebLock = Ptr
        self.AtlThunkSListPtr = Ptr
        self.IFEOKey = Ptr
        self.CrossProcessFlags = Ptr
        self.UserSharedInfoPtr = Ptr
        self.SystemReserved = ct.c_uint32
        self.AtlThunkSListPtr32 = ct.c_uint32
        self.ApiSetMap = Ptr
        self.TlsExpansionCounter = Ptr
        self.TlsBitmap = Ptr
        self.TlsBitmapBits = ct.c_uint32 * 2
        self.ReadOnlySharedMemoryBase = Ptr
        self.SharedData = Ptr # HotpatchInformation
        self.ReadOnlyStaticServerData = Ptr
        self.AnsiCodePageData = Ptr
        self.OemCodePageData = Ptr
        self.UnicodeCaseTableData = Ptr
        self.NumberOfProcessors = ct.c_uint32
        self.NtGlobalFlag = ct.c_uint32
        self.CriticalSectionTimeout = ct.c_longlong # LARGE_INTEGER
        self.HeapSegmentReserve = Ptr
        self.HeapSegmentCommit = Ptr
        self.HeapDeCommitTotalFreeThreshold = Ptr
        self.HeapDeCommitFreeBlockThreshold = Ptr
        self.NumberOfHeaps = ct.c_uint32
        self.MaximumNumberOfHeaps = ct.c_uint32
        self.ProcessHeaps = Ptr
        self.GdiSharedHandleTable = Ptr
        self.ProcessStarterHelper = Ptr
        self.GdiDCAttributeList = Ptr
        self.LoaderLock = Ptr
        self.OSMajorVersion = ct.c_uint32
        self.OSMinorVersion = ct.c_uint32
        self.OSBuildNumber = ct.c_uint16
        self.OSCSDVersion = ct.c_uint16
        self.OSPlatformId = ct.c_uint32
        self.ImageSubsystem = ct.c_uint32
        self.ImageSubsystemMajorVersion = ct.c_uint32
        self.ImageSubsystemMinorVersion = Ptr
        self.ActiveProcessAffinityMask = Ptr
        if ptr_size == 8:
            self.GdiHandleBuffer = ct.c_uint32 * 60
        else:
            self.GdiHandleBuffer = ct.c_uint32 * 34
        self.PostProcessInitRoutine = Ptr
        self.TlsExpansionBitmap = Ptr
        self.TlsExpansionBitmapBits = ct.c_uint32 * 32
        self.SessionId = Ptr
        self.AppCompatFlags = ct.c_ulonglong # ULARGE_INTEGER
        self.AppCompatFlagsUser = ct.c_ulonglong # ULARGE_INTEGER
        self.pShimData = Ptr
        self.AppCompatInfo = Ptr
        self.CSDVersion = UNICODE_STRING
        self.ActivationContextData = Ptr
        self.ProcessAssemblyStorageMap = Ptr
        self.SystemDefaultActivationContextData = Ptr
        self.SystemAssemblyStorageMap = Ptr
        self.MinimumStackCommit = Ptr
        self.FlsCallback = Ptr
        self.FlsListHead = LIST_ENTRY
        self.FlsBitmap = Ptr
        self.FlsBitmapBits = ct.c_uint32 * 4
        self.FlsHighIndex = Ptr
        self.WerRegistrationData = Ptr
        self.WerShipAssertPtr = Ptr
        self.pUnused = Ptr # pContextData
        self.pImageHeaderHash = Ptr
        self.TracingFlags = ct.c_uint64
        self.CsrServerReadOnlySharedMemoryBase = ct.c_uint64
        self.TppWorkerpListLock = Ptr
        self.TppWorkerpList = LIST_ENTRY
        self.WaitOnAddressHashTable = Ptr * 128


class PEB_LDR_DATA(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Length = ct.c_uint32
        self.Initialized = ct.c_uint8 * 4
        self.SsHandle = Ptr
        self.InLoadOrderModuleList = LIST_ENTRY
        self.InMemoryOrderModuleList = LIST_ENTRY
        self.InInitializationOrderModuleList = LIST_ENTRY
        self.EntryInProgress = Ptr
        self.ShutdownInProgress = ct.c_uint8
        self.ShutdownThreadId = Ptr


class LDR_DATA_TABLE_ENTRY(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.InLoadOrderLinks = LIST_ENTRY
        self.InMemoryOrderLinks = LIST_ENTRY
        self.InInitializationOrderLinks = LIST_ENTRY
        self.DllBase = Ptr
        self.EntryPoint = Ptr
        self.SizeOfImage = ct.c_uint32
        self.FullDllName = UNICODE_STRING
        self.BaseDllName = UNICODE_STRING
        self.Flags = ct.c_uint32
        self.LoadCount = ct.c_uint16


class LARGE_INTEGER(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.LowPart = ct.c_uint32
        self.HighPart = ct.c_uint32


class RTL_USER_PROCESS_PARAMETERS(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Reserved1 = ct.c_uint8 * 16
        self.Reserved2 = ct.c_uint32 * 10
        self.ImagePathName = UNICODE_STRING
        self.CommandLine = UNICODE_STRING



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import os
import io
import ntpath
import hashlib
import fnmatch
import shlex
import speakeasy.winenv.defs.windows.windows as windefs
import speakeasy.winenv.arch as _arch
from speakeasy.errors import FileSystemEmuError


def normalize_response_path(path):
    def _get_speakeasy_root():
        return os.path.join(os.path.dirname(__file__), os.pardir)

    root_var = '$ROOT$'
    if root_var in path:
        root = _get_speakeasy_root()
        return path.replace(root_var, root)
    return path


class MapView(object):
    """
    Represents a shared memory view
    """
    def __init__(self, base, offset, size, protect, process=None):
        self.base = base
        self.offset = offset
        self.size = size
        self.protect = protect
        self.process = process


class FileMap(object):
    """
    Represents a memory mapped file
    """
    curr_handle = 0x280

    def __init__(self, name, size, prot, backed_file=None):
        self.name = name
        self.backed_file = backed_file
        self.views = {}
        self.size = size

    def get_handle(self):
        hmap = FileMap.curr_handle
        FileMap.curr_handle += 4
        return hmap

    def get_name(self):
        return self.name

    def get_backed_file(self):
        return self.backed_file

    def add_view(self, base, offset, size, protect):
        view = MapView(base, offset, size, protect)
        self.views.update({base: view})


class File(object):
    """
    Base class for an emulated file
    """
    curr_handle = 0x80

    def __init__(self, path, config={}, data=b''):
        self.path = path
        self.data = None
        self.bytes_written = 0
        if data:
            self.data = io.BytesIO(data)
        self.curr_offset = 0
        self.is_dir = False
        self.config = config

    def duplicate(self):
        new = File(self.path, config=self.config, data=self.data.getvalue())
        new.is_dir = self.is_dir
        return new
        
    def get_handle(self):
        hfile = File.curr_handle
        File.curr_handle += 4
        return hfile

    def get_path(self):
        return self.path

    def get_hash(self):
        h = hashlib.sha256()
        data = self.get_data(reset_pointer=True)
        h.update(data)
        return h.hexdigest()

    def get_size(self):
        if not self.data and self.config:
            self.data = self.handle_file_data()
        if not self.data:
            return 0
        off = self.data.tell()
        self.data.seek(0, io.SEEK_SET)
        size = len(self.data.read())
        self.data.seek(off, io.SEEK_SET)
        return size

    def get_data(self, size=-1, reset_pointer=False):
        if not self.data and self.config:
            self.data = self.handle_file_data()

        if not self.data:
            return b''

        off = self.data.tell()
        if off == self.get_size():
            if reset_pointer:
                # Reset the file pointer
                self.data.seek(0)
            else:
                return b''

        return self.data.read(size)

    def seek(self, offset, whence):
        if whence not in [io.SEEK_CUR, io.SEEK_SET, io.SEEK_END]:
            return
        if self.data:
            self.data.seek(offset, whence)

    def tell(self):
        if self.data:
            return self.data.tell()
        return None

    def add_data(self, data):

        if not self.data:
            self.data = io.BytesIO()
        off = self.data.tell()
        self.data.seek(0, io.SEEK_END)
        self.data.write(data)
        self.data.seek(off, io.SEEK_SET)
        self.bytes_written += len(data)

    def remove_data(self):
        self.data = io.BytesIO(b'')

    def is_directory(self):
        return self.is_dir

    def handle_file_data(self):
        """
        Based on the emulation config, determine what data
        to return from the read request
        """

        path = self.config.get('path')
        if path:
            path = normalize_response_path(path)
            with open(path, 'rb') as f:
                return io.BytesIO(f.read())
        bf = self.config.get('byte_fill')
        if bf:
            byte = bf.get('byte')
            if byte.startswith('0x'):
                byte = 0xFF & int(byte, 0)
            else:
                byte = 0xFF & int(byte, 16)
            size = bf.get('size')
            b = (byte).to_bytes(1, 'little')
            return b * size
        return io.BytesIO(b'')


class Pipe(File):
    """
    Emulated named pipe objects
    """
    curr_handle = 0x400

    def __init__(self, name, mode, num_instances, out_size, in_size, config={}):
        super(Pipe, self).__init__(path=name, config=config)
        self.name = name
        self.mode = mode
        self.num_instances = num_instances
        self.out_size = out_size
        self.in_size = in_size

    def get_handle(self):
        hpipe = Pipe.curr_handle
        Pipe.curr_handle += 4
        return hpipe


class FileManager(object):
    """
    Manages file system activity during emulation
    """
    def __init__(self, config, emu):
        super(FileManager, self).__init__()
        self.file_handles = {}
        self.pipe_handles = {}
        self.file_maps = {}

        # top level config
        self.config = config

        # "files" key of config
        self.file_config = self.config.get('filesystem', {})
        self.emu = emu

        cmdline = self.config.get('command_line')

        if cmdline is None:
            cmdline = ""

        self.emulated_binname = shlex.split(cmdline)[0]

        # First file in this list seems to always be the module itself
        self.files = []

    def file_create_mapping(self, hfile, name, size, prot):
        if hfile not in (windefs.INVALID_HANDLE_VALUE, 0):
            f = self.get_file_from_handle(hfile)
            fm = FileMap(name, size, prot, f)
            hnd = fm.get_handle()
            self.file_maps.update({hnd: fm})
            return hnd
        else:
            fm = FileMap(name, size, prot, None)
            hnd = fm.get_handle()
            self.file_maps.update({hnd: fm})
            return hnd

    def walk_files(self):
        for f in self.file_config.get('files', []):
            path = f.get('emu_path')
            if not path:
                continue
            yield path

    def get_dropped_files(self):
        return [f for f in self.files if f.bytes_written == f.get_size()]

    def get_mapping_from_handle(self, handle):
        return self.file_maps.get(handle)

    def get_mapping_from_addr(self, addr):
        for h, fmap in self.file_maps.items():
            for base, view in fmap.views.items():
                if base == addr:
                    return fmap

    def get_file_from_handle(self, handle):
        return self.file_handles.get(handle)

    def get_pipe_from_handle(self, handle):
        return self.pipe_handles.get(handle)

    def get_file_from_path(self, path):
        # The emulated sample is requesting itself. The module path
        # for it is(?) always the first entry in self.files
        if self.emulated_binname in path:
            return self.files[0]

        for f in self.files:
            if f.get_path().lower() == path.lower():
                return f
        return None

    def get_all_files(self):
        return self.files

    def handle_file_data(self, fconf):

        path = fconf.get('path')
        if path:
            path = normalize_response_path(path)
            with open(path, 'rb') as f:
                return f.read()
        bf = fconf.get('byte_fill')
        if bf:
            byte = bf.get('byte')
            if byte.startswith('0x'):
                byte = 0xFF & int(byte, 0)
            else:
                byte = 0xFF & int(byte, 16)
            size = bf.get('size')
            b = (byte).to_bytes(1, 'little')
            return b * size

    def add_existing_file(self, path, data):
        """
        Register an existing file already included in the emulation space
        (with data)
        """
        f = File(path, data=data)
        self.files.append(f)
        return f

    def create_file(self, path):
        f = self.get_file_from_path(path)
        if f:
            self.files.remove(f)
        f = File(path)
        self.files.append(f)
        return f

    def delete_file(self, path):
        f = self.get_file_from_path(path)
        if f:
            self.files.remove(f)
            return True
        return False

    def get_emu_file(self, path):
        # Does this file exist in our emulation environment
        # See if we have a handler for this exact file
        for f in self.file_config.get('files', []):
            mode = f.get('mode')
            if mode == 'full_path':
                if fnmatch.fnmatch(path.lower(), f.get('emu_path').lower()):
                    return f

        all_modules = self.config.get('modules')

        if self.emu.arch == _arch.ARCH_X86:
            decoy_dir = all_modules.get('module_directory_x86', [])
        else:
            decoy_dir = all_modules.get('module_directory_x64', [])

        ext = os.path.splitext(path)[1]

        # Check if we can load the contents of a decoy DLL
        for f in all_modules.get('user_modules', []):
            if f.get('path') == path:
                newconf = dict()
                newconf['path'] = os.path.join(decoy_dir, f.get('name') + ext)
                return newconf

        for f in all_modules.get('system_modules', []):
            if f.get('path') == path:
                newconf = dict()
                newconf['path'] = os.path.join(decoy_dir, f.get('name') + ext)
                return newconf

        # If no full path handler exists, do we have an extension handler?
        for f in self.file_config.get('files', []):
            path_ext = ntpath.splitext(path)[-1:][0].strip('.')
            if path_ext:
                mode = f.get('mode')
                if mode == 'by_ext':
                    if path_ext.lower() == f.get('ext'):
                        return f

        # Finally, do we have a catch-all default handler?
        for f in self.file_config.get('files', []):

            mode = f.get('mode')
            if mode == 'default':
                return f
        return None

    def pipe_open(self, path, mode, num_instances, out_size, in_size):
        hnd = None
        fconf = self.get_emu_file(path)
        if not fconf:
            return hnd
        p = Pipe(path, mode, num_instances, out_size, in_size, config=fconf)
        hnd = p.get_handle()
        self.pipe_handles.update({hnd: p})
        return hnd

    def does_file_exist(self, path):
        if self.get_file_from_path(path):
            return True

        if self.get_emu_file(path):
            return True
        return False

    def get_object_from_handle(self, handle):
        obj = self.file_maps.get(handle)
        if obj:
            return obj
        obj = self.pipe_handles.get(handle)
        if obj:
            return obj
        obj = self.file_handles.get(handle)
        if obj:
            return obj

    def file_open(self, path, create=False, truncate=False, is_dir=False):
        hnd = None

        if create:
            f = self.create_file(path)
            hnd = f.get_handle()
            self.file_handles.update({hnd: f})
        else:
            f = self.get_file_from_path(path)

            if f:
                # Deep-copy this file so we can have separate file
                # offset pointers
                dup = f.duplicate()
                hnd = dup.get_handle()
                self.file_handles.update({hnd: dup})
                return hnd

            fconf = self.get_emu_file(path)
            if not fconf:
                return hnd

            real_path = fconf.get('path', '')
            real_path = normalize_response_path(real_path)
            if not truncate:
                if real_path and not os.path.exists(real_path):
                    raise FileSystemEmuError('File path not found: %s' % (real_path))
                f = File(path, config=fconf)
                self.files.append(f)
            else:
                if real_path and not os.path.exists(real_path):
                    raise FileSystemEmuError('File path not found: %s' % (real_path))
                f = File(path, config=fconf)
                self.files.append(f)
            hnd = f.get_handle()
            self.file_handles.update({hnd: f})

        return hnd
